@@ -45,6 +45,8 @@
 #include <infiniband/acm_prov.h>
 #include <infiniband/umad.h>
 #include <infiniband/verbs.h>
+#include <infiniband/umad_sa.h>
+#include <infiniband/umad_sa_mcm.h>
 #include <ifaddrs.h>
 #include <dlfcn.h>
 #include <search.h>
@@ -60,10 +62,6 @@
 #include "acm_util.h"
 #include "acm_mad.h"
 
-#define src_out     data[0]
-#define src_index   data[1]
-#define dst_index   data[2]
-
 #define IB_LID_MCAST_START 0xc000
 
 #define MAX_EP_ADDR 4
@@ -361,8 +359,10 @@ static void
 acmp_remove_dest(struct acmp_ep *ep, struct acmp_dest *dest)
 {
 	acm_log(2, "%s\n", dest->name);
-	tdelete(dest->address, &ep->dest_map[dest->addr_type - 1],
-		acmp_compare_dest);
+	if (!tdelete(dest->address, &ep->dest_map[dest->addr_type - 1],
+		     acmp_compare_dest))
+		acm_log(0, "ERROR: %s not found!!\n", dest->name);
+
 	acmp_put_dest(dest);
 }
 
@@ -1378,15 +1378,16 @@ static void acmp_init_join(struct ib_sa_mad *mad, union ibv_gid *port_gid,
 		IB_COMP_MASK_MC_SCOPE | IB_COMP_MASK_MC_JOIN_STATE;
 
 	mc_rec = (struct ib_mc_member_rec *) mad->data;
-	acmp_format_mgid(&mc_rec->mgid, pkey | 0x8000, tos, rate, mtu);
+	acmp_format_mgid(&mc_rec->mgid, pkey | IB_PKEY_FULL_MEMBER, tos, rate, mtu);
 	mc_rec->port_gid = *port_gid;
 	mc_rec->qkey = htobe32(ACM_QKEY);
-	mc_rec->mtu = 0x80 | mtu;
+	mc_rec->mtu = umad_sa_set_rate_mtu_or_life(UMAD_SA_SELECTOR_EXACTLY, mtu);
 	mc_rec->tclass = tclass;
 	mc_rec->pkey = htobe16(pkey);
-	mc_rec->rate = 0x80 | rate;
-	mc_rec->sl_flow_hop = htobe32(((uint32_t) sl) << 28);
-	mc_rec->scope_state = 0x51;
+	mc_rec->rate = umad_sa_set_rate_mtu_or_life(UMAD_SA_SELECTOR_EXACTLY, rate);
+	mc_rec->sl_flow_hop = umad_sa_mcm_set_sl_flow_hop(sl, 0, 0);
+	mc_rec->scope_state = umad_sa_mcm_set_scope_state(UMAD_SA_MCM_ADDR_SCOPE_SITE_LOCAL,
+							  UMAD_SA_MCM_JOIN_STATE_FULL_MEMBER);
 }
 
 static void acmp_join_group(struct acmp_ep *ep, union ibv_gid *port_gid,
@@ -2412,8 +2413,43 @@ static int acmp_add_addr(const struct acm_address *addr, void *ep_context,
 static void acmp_remove_addr(void *addr_context)
 {
 	struct acmp_addr *address = addr_context;
+	struct acmp_device *dev;
+	struct acmp_dest *dest;
+	struct acmp_ep *ep;
+	int i;
 
 	acm_log(2, "\n");
+
+	/*
+	 * The address may be a local destination address. If so,
+	 * delete it from the cache.
+	 */
+
+	pthread_mutex_lock(&acmp_dev_lock);
+	list_for_each(&acmp_dev_list, dev, entry) {
+		pthread_mutex_unlock(&acmp_dev_lock);
+
+		for (i = 0; i < dev->port_cnt; i++) {
+			struct acmp_port *port = &dev->port[i];
+
+			pthread_mutex_lock(&port->lock);
+			list_for_each(&port->ep_list, ep, entry) {
+				pthread_mutex_unlock(&port->lock);
+				dest = acmp_get_dest(ep, address->type, address->addr->info.addr);
+				if (dest) {
+					acm_log(2, "Found a dest addr, deleting it\n");
+					pthread_mutex_lock(&ep->lock);
+					acmp_remove_dest(ep, dest);
+					pthread_mutex_unlock(&ep->lock);
+				}
+				pthread_mutex_lock(&port->lock);
+			}
+			pthread_mutex_unlock(&port->lock);
+		}
+		pthread_mutex_lock(&acmp_dev_lock);
+	}
+	pthread_mutex_unlock(&acmp_dev_lock);
+
 	memset(address, 0, sizeof(*address));
 }
 
@@ -2453,20 +2489,16 @@ acmp_get_ep(struct acmp_port *port, struct acm_endpoint *endpoint)
 static uint16_t acmp_get_pkey_index(struct acm_endpoint *endpoint)
 {
 	struct acmp_port *port;
-	int ret;
-	__be16 pkey;
-	uint16_t i;
+	int i;
 
 	port = acmp_get_port(endpoint);
 	if (!port)
 		return 0;
-
-	for (i = 0, ret = 0; !ret; i++) {
-		ret = ibv_query_pkey(port->dev->verbs, port->port_num, i, &pkey);
-		if (!ret && endpoint->pkey == be16toh(pkey))
-			return i;
-	}
-	return 0;
+	i = ibv_get_pkey_index(port->dev->verbs, port->port_num,
+			       htobe16(endpoint->pkey));
+	if (i < 0)
+		return 0;
+	return i;
 }
 
 static void acmp_close_endpoint(void *ep_context)
@@ -2622,6 +2654,7 @@ static void acmp_port_up(struct acmp_port *port)
 	__be16 pkey_be;
 	__be16 sm_lid;
 	int i, ret;
+	int instance;
 
 	acm_log(1, "%s %d\n", port->dev->verbs->device->name, port->port_num);
 	ret = ibv_query_port(port->dev->verbs, port->port_num, &attr);
@@ -2647,7 +2680,7 @@ static void acmp_port_up(struct acmp_port *port)
 	acmp_set_dest_addr(&port->sa_dest, ACM_ADDRESS_LID,
 			   (uint8_t *) &sm_lid, sizeof(sm_lid));
 
-	atomic_set(&port->sa_dest.refcnt, 1);
+	instance = atomic_inc(&port->sa_dest.refcnt) - 1;
 	port->sa_dest.state = ACMP_READY;
 	for (i = 0; i < attr.pkey_tbl_len; i++) {
 		ret = ibv_query_pkey(port->dev->verbs, port->port_num, i, &pkey_be);
@@ -2667,11 +2700,13 @@ static void acmp_port_up(struct acmp_port *port)
 	}
 
 	port->state = IBV_PORT_ACTIVE;
-	acm_log(1, "%s %d is up\n", port->dev->verbs->device->name, port->port_num);
+	acm_log(1, "%s %d %d is up\n", port->dev->verbs->device->name, port->port_num, instance);
 }
 
 static void acmp_port_down(struct acmp_port *port)
 {
+	int instance;
+
 	acm_log(1, "%s %d\n", port->dev->verbs->device->name, port->port_num);
 	pthread_mutex_lock(&port->lock);
 	port->state = IBV_PORT_DOWN;
@@ -2682,13 +2717,13 @@ static void acmp_port_down(struct acmp_port *port)
 	 * event instead of a sleep loop, but it's not worth it given how
 	 * infrequently we should be processing a port down event in practice.
 	 */
-	atomic_dec(&port->sa_dest.refcnt);
-	while (atomic_get(&port->sa_dest.refcnt))
-		sleep(0);
-	pthread_mutex_lock(&port->sa_dest.lock);
-	port->sa_dest.state = ACMP_INIT;
-	pthread_mutex_unlock(&port->sa_dest.lock);
-	acm_log(1, "%s %d is down\n", port->dev->verbs->device->name, port->port_num);
+	instance = atomic_dec(&port->sa_dest.refcnt);
+	if (instance == 1) {
+		pthread_mutex_lock(&port->sa_dest.lock);
+		port->sa_dest.state = ACMP_INIT;
+		pthread_mutex_unlock(&port->sa_dest.lock);
+	}
+	acm_log(1, "%s %d %d is down\n", port->dev->verbs->device->name, port->port_num, instance);
 }
 
 static int acmp_open_port(const struct acm_port *cport, void *dev_context,
