@@ -46,6 +46,7 @@
 
 #include <util/compiler.h>
 #include <util/symver.h>
+#include <infiniband/cmd_write.h>
 
 #include "ibverbs.h"
 #ifndef NRESOLVE_NEIGH
@@ -68,6 +69,10 @@ int __attribute__((const)) ibv_rate_to_mult(enum ibv_rate rate)
 	case IBV_RATE_60_GBPS:  return 24;
 	case IBV_RATE_80_GBPS:  return 32;
 	case IBV_RATE_120_GBPS: return 48;
+	case IBV_RATE_28_GBPS:  return 11;
+	case IBV_RATE_50_GBPS:  return 20;
+	case IBV_RATE_400_GBPS: return 160;
+	case IBV_RATE_600_GBPS: return 240;
 	default:           return -1;
 	}
 }
@@ -84,6 +89,10 @@ enum ibv_rate __attribute__((const)) mult_to_ibv_rate(int mult)
 	case 24: return IBV_RATE_60_GBPS;
 	case 32: return IBV_RATE_80_GBPS;
 	case 48: return IBV_RATE_120_GBPS;
+	case 11: return IBV_RATE_28_GBPS;
+	case 20: return IBV_RATE_50_GBPS;
+	case 160: return IBV_RATE_400_GBPS;
+	case 240: return IBV_RATE_600_GBPS;
 	default: return IBV_RATE_MAX;
 	}
 }
@@ -108,6 +117,10 @@ int  __attribute__((const)) ibv_rate_to_mbps(enum ibv_rate rate)
 	case IBV_RATE_100_GBPS: return 103125;
 	case IBV_RATE_200_GBPS: return 206250;
 	case IBV_RATE_300_GBPS: return 309375;
+	case IBV_RATE_28_GBPS:  return 28125;
+	case IBV_RATE_50_GBPS:  return 53125;
+	case IBV_RATE_400_GBPS: return 425000;
+	case IBV_RATE_600_GBPS: return 637500;
 	default:               return -1;
 	}
 }
@@ -132,6 +145,10 @@ enum ibv_rate __attribute__((const)) mbps_to_ibv_rate(int mbps)
 	case 103125: return IBV_RATE_100_GBPS;
 	case 206250: return IBV_RATE_200_GBPS;
 	case 309375: return IBV_RATE_300_GBPS;
+	case 28125:  return IBV_RATE_28_GBPS;
+	case 53125:  return IBV_RATE_50_GBPS;
+	case 425000: return IBV_RATE_400_GBPS;
+	case 637500: return IBV_RATE_600_GBPS;
 	default:     return IBV_RATE_MAX;
 	}
 }
@@ -141,15 +158,64 @@ LATEST_SYMVER_FUNC(ibv_query_device, 1_1, "IBVERBS_1.1",
 		   struct ibv_context *context,
 		   struct ibv_device_attr *device_attr)
 {
-	return context->ops.query_device(context, device_attr);
+	return get_ops(context)->query_device(context, device_attr);
 }
 
+int __lib_query_port(struct ibv_context *context, uint8_t port_num,
+		     struct ibv_port_attr *port_attr, size_t port_attr_len)
+{
+	/* Don't expose this mess to the provider, provide a large enough
+	 * temporary buffer if the user buffer is too small.
+	 */
+	if (port_attr_len < sizeof(struct ibv_port_attr)) {
+		struct ibv_port_attr tmp_attr = {};
+		int rc;
+
+		rc = get_ops(context)->query_port(context, port_num,
+						    &tmp_attr);
+		if (rc)
+			return rc;
+
+		memcpy(port_attr, &tmp_attr, port_attr_len);
+		return 0;
+	}
+
+	memset(port_attr, 0, port_attr_len);
+	return get_ops(context)->query_port(context, port_num, port_attr);
+}
+
+struct _compat_ibv_port_attr {
+	enum ibv_port_state state;
+	enum ibv_mtu max_mtu;
+	enum ibv_mtu active_mtu;
+	int gid_tbl_len;
+	uint32_t port_cap_flags;
+	uint32_t max_msg_sz;
+	uint32_t bad_pkey_cntr;
+	uint32_t qkey_viol_cntr;
+	uint16_t pkey_tbl_len;
+	uint16_t lid;
+	uint16_t sm_lid;
+	uint8_t lmc;
+	uint8_t max_vl_num;
+	uint8_t sm_sl;
+	uint8_t subnet_timeout;
+	uint8_t init_type_reply;
+	uint8_t active_width;
+	uint8_t active_speed;
+	uint8_t phys_state;
+	uint8_t link_layer;
+	uint8_t flags;
+};
+
 LATEST_SYMVER_FUNC(ibv_query_port, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_context *context, uint8_t port_num,
-		   struct ibv_port_attr *port_attr)
+		   struct _compat_ibv_port_attr *port_attr)
 {
-	return context->ops.query_port(context, port_num, port_attr);
+	return __lib_query_port(context, port_num,
+				(struct ibv_port_attr *)port_attr,
+				sizeof(*port_attr));
 }
 
 LATEST_SYMVER_FUNC(ibv_query_gid, 1_1, "IBVERBS_1.1",
@@ -200,13 +266,29 @@ LATEST_SYMVER_FUNC(ibv_query_pkey, 1_1, "IBVERBS_1.1",
 	return 0;
 }
 
+LATEST_SYMVER_FUNC(ibv_get_pkey_index, 1_5, "IBVERBS_1.5",
+		   int,
+		   struct ibv_context *context, uint8_t port_num, __be16 pkey)
+{
+	__be16 pkey_i;
+	int i, ret;
+
+	for (i = 0; ; i++) {
+		ret = ibv_query_pkey(context, port_num, i, &pkey_i);
+		if (ret < 0)
+			return ret;
+		if (pkey == pkey_i)
+			return i;
+	}
+}
+
 LATEST_SYMVER_FUNC(ibv_alloc_pd, 1_1, "IBVERBS_1.1",
 		   struct ibv_pd *,
 		   struct ibv_context *context)
 {
 	struct ibv_pd *pd;
 
-	pd = context->ops.alloc_pd(context);
+	pd = get_ops(context)->alloc_pd(context);
 	if (pd)
 		pd->context = context;
 
@@ -217,7 +299,7 @@ LATEST_SYMVER_FUNC(ibv_dealloc_pd, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_pd *pd)
 {
-	return pd->context->ops.dealloc_pd(pd);
+	return get_ops(pd->context)->dealloc_pd(pd);
 }
 
 LATEST_SYMVER_FUNC(ibv_reg_mr, 1_1, "IBVERBS_1.1",
@@ -230,7 +312,7 @@ LATEST_SYMVER_FUNC(ibv_reg_mr, 1_1, "IBVERBS_1.1",
 	if (ibv_dontfork_range(addr, length))
 		return NULL;
 
-	mr = pd->context->ops.reg_mr(pd, addr, length, access);
+	mr = get_ops(pd->context)->reg_mr(pd, addr, length, access);
 	if (mr) {
 		mr->context = pd->context;
 		mr->pd      = pd;
@@ -253,6 +335,11 @@ LATEST_SYMVER_FUNC(ibv_rereg_mr, 1_1, "IBVERBS_1.1",
 	void *old_addr;
 	size_t old_len;
 
+	if (verbs_get_mr(mr)->mr_type != IBV_MR_TYPE_MR) {
+		errno = EINVAL;
+		return IBV_REREG_MR_ERR_INPUT;
+	}
+
 	if (flags & ~IBV_REREG_MR_FLAGS_SUPPORTED) {
 		errno = EINVAL;
 		return IBV_REREG_MR_ERR_INPUT;
@@ -278,7 +365,9 @@ LATEST_SYMVER_FUNC(ibv_rereg_mr, 1_1, "IBVERBS_1.1",
 
 	old_addr = mr->addr;
 	old_len = mr->length;
-	err = mr->context->ops.rereg_mr(mr, flags, pd, addr, length, access);
+	err = get_ops(mr->context)->rereg_mr(verbs_get_mr(mr),
+					     flags, pd, addr,
+					     length, access);
 	if (!err) {
 		if (flags & IBV_REREG_MR_CHANGE_PD)
 			mr->pd = pd;
@@ -305,11 +394,12 @@ LATEST_SYMVER_FUNC(ibv_dereg_mr, 1_1, "IBVERBS_1.1",
 		   struct ibv_mr *mr)
 {
 	int ret;
-	void *addr	= mr->addr;
-	size_t length	= mr->length;
+	void *addr		= mr->addr;
+	size_t length		= mr->length;
+	enum ibv_mr_type type	= verbs_get_mr(mr)->mr_type;
 
-	ret = mr->context->ops.dereg_mr(mr);
-	if (!ret)
+	ret = get_ops(mr->context)->dereg_mr(verbs_get_mr(mr));
+	if (!ret && type == IBV_MR_TYPE_MR)
 		ibv_dofork_range(addr, length);
 
 	return ret;
@@ -317,22 +407,21 @@ LATEST_SYMVER_FUNC(ibv_dereg_mr, 1_1, "IBVERBS_1.1",
 
 struct ibv_comp_channel *ibv_create_comp_channel(struct ibv_context *context)
 {
-	struct ibv_comp_channel            *channel;
-	struct ibv_create_comp_channel      cmd;
+	struct ibv_create_comp_channel req;
 	struct ib_uverbs_create_comp_channel_resp resp;
+	struct ibv_comp_channel            *channel;
 
 	channel = malloc(sizeof *channel);
 	if (!channel)
 		return NULL;
 
-	IBV_INIT_CMD_RESP(&cmd, sizeof cmd, CREATE_COMP_CHANNEL, &resp, sizeof resp);
-	if (write(context->cmd_fd, &cmd, sizeof cmd) != sizeof cmd) {
+	req.core_payload = (struct ib_uverbs_create_comp_channel){};
+	if (execute_cmd_write(context, IB_USER_VERBS_CMD_CREATE_COMP_CHANNEL,
+			      &req, sizeof(req), &resp, sizeof(resp))) {
 		free(channel);
 		return NULL;
 	}
 
-	(void) VALGRIND_MAKE_MEM_DEFINED(&resp, sizeof resp);
-
 	channel->context = context;
 	channel->fd      = resp.fd;
 	channel->refcnt  = 0;
@@ -370,7 +459,7 @@ LATEST_SYMVER_FUNC(ibv_create_cq, 1_1, "IBVERBS_1.1",
 {
 	struct ibv_cq *cq;
 
-	cq = context->ops.create_cq(context, cqe, channel, comp_vector);
+	cq = get_ops(context)->create_cq(context, cqe, channel, comp_vector);
 
 	if (cq)
 		verbs_init_cq(cq, context, channel, cq_context);
@@ -382,7 +471,7 @@ LATEST_SYMVER_FUNC(ibv_resize_cq, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_cq *cq, int cqe)
 {
-	return cq->context->ops.resize_cq(cq, cqe);
+	return get_ops(cq->context)->resize_cq(cq, cqe);
 }
 
 LATEST_SYMVER_FUNC(ibv_destroy_cq, 1_1, "IBVERBS_1.1",
@@ -392,7 +481,7 @@ LATEST_SYMVER_FUNC(ibv_destroy_cq, 1_1, "IBVERBS_1.1",
 	struct ibv_comp_channel *channel = cq->channel;
 	int ret;
 
-	ret = cq->context->ops.destroy_cq(cq);
+	ret = get_ops(cq->context)->destroy_cq(cq);
 
 	if (channel) {
 		if (!ret) {
@@ -418,7 +507,7 @@ LATEST_SYMVER_FUNC(ibv_get_cq_event, 1_1, "IBVERBS_1.1",
 	*cq         = (struct ibv_cq *) (uintptr_t) ev.cq_handle;
 	*cq_context = (*cq)->cq_context;
 
-	(*cq)->context->ops.cq_event(*cq);
+	get_ops((*cq)->context)->cq_event(*cq);
 
 	return 0;
 }
@@ -440,7 +529,7 @@ LATEST_SYMVER_FUNC(ibv_create_srq, 1_1, "IBVERBS_1.1",
 {
 	struct ibv_srq *srq;
 
-	srq = pd->context->ops.create_srq(pd, srq_init_attr);
+	srq = get_ops(pd->context)->create_srq(pd, srq_init_attr);
 	if (srq) {
 		srq->context          = pd->context;
 		srq->srq_context      = srq_init_attr->srq_context;
@@ -459,21 +548,21 @@ LATEST_SYMVER_FUNC(ibv_modify_srq, 1_1, "IBVERBS_1.1",
 		   struct ibv_srq_attr *srq_attr,
 		   int srq_attr_mask)
 {
-	return srq->context->ops.modify_srq(srq, srq_attr, srq_attr_mask);
+	return get_ops(srq->context)->modify_srq(srq, srq_attr, srq_attr_mask);
 }
 
 LATEST_SYMVER_FUNC(ibv_query_srq, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_srq *srq, struct ibv_srq_attr *srq_attr)
 {
-	return srq->context->ops.query_srq(srq, srq_attr);
+	return get_ops(srq->context)->query_srq(srq, srq_attr);
 }
 
 LATEST_SYMVER_FUNC(ibv_destroy_srq, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_srq *srq)
 {
-	return srq->context->ops.destroy_srq(srq);
+	return get_ops(srq->context)->destroy_srq(srq);
 }
 
 LATEST_SYMVER_FUNC(ibv_create_qp, 1_1, "IBVERBS_1.1",
@@ -481,7 +570,7 @@ LATEST_SYMVER_FUNC(ibv_create_qp, 1_1, "IBVERBS_1.1",
 		   struct ibv_pd *pd,
 		   struct ibv_qp_init_attr *qp_init_attr)
 {
-	struct ibv_qp *qp = pd->context->ops.create_qp(pd, qp_init_attr);
+	struct ibv_qp *qp = get_ops(pd->context)->create_qp(pd, qp_init_attr);
 
 	if (qp) {
 		qp->context    	     = pd->context;
@@ -500,6 +589,15 @@ LATEST_SYMVER_FUNC(ibv_create_qp, 1_1, "IBVERBS_1.1",
 	return qp;
 }
 
+struct ibv_qp_ex *ibv_qp_to_qp_ex(struct ibv_qp *qp)
+{
+	struct verbs_qp *vqp = (struct verbs_qp *)qp;
+
+	if (vqp->comp_mask & VERBS_QP_EX)
+		return &vqp->qp_ex;
+	return NULL;
+}
+
 LATEST_SYMVER_FUNC(ibv_query_qp, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_qp *qp, struct ibv_qp_attr *attr,
@@ -508,7 +606,7 @@ LATEST_SYMVER_FUNC(ibv_query_qp, 1_1, "IBVERBS_1.1",
 {
 	int ret;
 
-	ret = qp->context->ops.query_qp(qp, attr, attr_mask, init_attr);
+	ret = get_ops(qp->context)->query_qp(qp, attr, attr_mask, init_attr);
 	if (ret)
 		return ret;
 
@@ -525,7 +623,7 @@ LATEST_SYMVER_FUNC(ibv_modify_qp, 1_1, "IBVERBS_1.1",
 {
 	int ret;
 
-	ret = qp->context->ops.modify_qp(qp, attr, attr_mask);
+	ret = get_ops(qp->context)->modify_qp(qp, attr, attr_mask);
 	if (ret)
 		return ret;
 
@@ -539,14 +637,14 @@ LATEST_SYMVER_FUNC(ibv_destroy_qp, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_qp *qp)
 {
-	return qp->context->ops.destroy_qp(qp);
+	return get_ops(qp->context)->destroy_qp(qp);
 }
 
 LATEST_SYMVER_FUNC(ibv_create_ah, 1_1, "IBVERBS_1.1",
 		   struct ibv_ah *,
 		   struct ibv_pd *pd, struct ibv_ah_attr *attr)
 {
-	struct ibv_ah *ah = pd->context->ops.create_ah(pd, attr);
+	struct ibv_ah *ah = get_ops(pd->context)->create_ah(pd, attr);
 
 	if (ah) {
 		ah->context = pd->context;
@@ -819,21 +917,21 @@ LATEST_SYMVER_FUNC(ibv_destroy_ah, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_ah *ah)
 {
-	return ah->context->ops.destroy_ah(ah);
+	return get_ops(ah->context)->destroy_ah(ah);
 }
 
 LATEST_SYMVER_FUNC(ibv_attach_mcast, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_qp *qp, const union ibv_gid *gid, uint16_t lid)
 {
-	return qp->context->ops.attach_mcast(qp, gid, lid);
+	return get_ops(qp->context)->attach_mcast(qp, gid, lid);
 }
 
 LATEST_SYMVER_FUNC(ibv_detach_mcast, 1_1, "IBVERBS_1.1",
 		   int,
 		   struct ibv_qp *qp, const union ibv_gid *gid, uint16_t lid)
 {
-	return qp->context->ops.detach_mcast(qp, gid, lid);
+	return get_ops(qp->context)->detach_mcast(qp, gid, lid);
 }
 
 static inline int ipv6_addr_v4mapped(const struct in6_addr *a)
@@ -932,10 +1030,12 @@ int ibv_resolve_eth_l2_from_gid(struct ibv_context *context,
 	if (process_get_neigh(&neigh_handler))
 		goto free_resources;
 
-	ret_vid = neigh_get_vlan_id_from_dev(&neigh_handler);
+	if (vid) {
+		ret_vid = neigh_get_vlan_id_from_dev(&neigh_handler);
 
-	if (ret_vid <= 0xfff)
-		neigh_set_vlan_id(&neigh_handler, ret_vid);
+		if (ret_vid <= 0xfff)
+			neigh_set_vlan_id(&neigh_handler, ret_vid);
+	}
 
 	/* We are using only Ethernet here */
 	ether_len = neigh_get_ll(&neigh_handler,
@@ -945,7 +1045,8 @@ int ibv_resolve_eth_l2_from_gid(struct ibv_context *context,
 	if (ether_len <= 0)
 		goto free_resources;
 
-	*vid = ret_vid;
+	if (vid)
+		*vid = ret_vid;
 
 	ret = 0;
 
