@@ -538,8 +538,10 @@ static int rdma_create_id2(struct rdma_event_channel *channel,
 	cmd.qp_type = qp_type;
 
 	ret = write(id_priv->id.channel->fd, &cmd, sizeof cmd);
-	if (ret != sizeof cmd)
+	if (ret != sizeof(cmd)) {
+		ret = (ret >= 0) ? ERR(ENODATA) : -1;
 		goto err;
+	}
 
 	VALGRIND_MAKE_MEM_DEFINED(&resp, sizeof resp);
 
@@ -989,8 +991,8 @@ static int ucma_is_ud_qp(enum ibv_qp_type qp_type)
 	return (qp_type == IBV_QPT_UD);
 }
 
-static int rdma_init_qp_attr(struct rdma_cm_id *id, struct ibv_qp_attr *qp_attr,
-			     int *qp_attr_mask)
+int rdma_init_qp_attr(struct rdma_cm_id *id, struct ibv_qp_attr *qp_attr,
+		      int *qp_attr_mask)
 {
 	struct ucma_abi_init_qp_attr cmd;
 	struct ib_uverbs_qp_attr resp;
@@ -1021,7 +1023,7 @@ static int ucma_modify_qp_rtr(struct rdma_cm_id *id, uint8_t resp_res)
 	uint8_t link_layer;
 
 	if (!id->qp)
-		return ERR(EINVAL);
+		return 0;
 
 	/* Need to update QP attributes from default values. */
 	qp_attr.qp_state = IBV_QPS_INIT;
@@ -1058,6 +1060,9 @@ static int ucma_modify_qp_rts(struct rdma_cm_id *id, uint8_t init_depth)
 	struct ibv_qp_attr qp_attr;
 	int qp_attr_mask, ret;
 
+	if (!id->qp)
+		return 0;
+
 	qp_attr.qp_state = IBV_QPS_RTS;
 	ret = rdma_init_qp_attr(id, &qp_attr, &qp_attr_mask);
 	if (ret)
@@ -1090,33 +1095,17 @@ static int ucma_modify_qp_err(struct rdma_cm_id *id)
 	return rdma_seterrno(ibv_modify_qp(id->qp, &qp_attr, IBV_QP_STATE));
 }
 
-static int ucma_find_pkey(struct cma_device *cma_dev, uint8_t port_num,
-			  __be16 pkey, uint16_t *pkey_index)
-{
-	int ret, i;
-	__be16 chk_pkey;
-
-	for (i = 0, ret = 0; !ret; i++) {
-		ret = ibv_query_pkey(cma_dev->verbs, port_num, i, &chk_pkey);
-		if (!ret && pkey == chk_pkey) {
-			*pkey_index = (uint16_t) i;
-			return 0;
-		}
-	}
-	return ERR(EINVAL);
-}
-
 static int ucma_init_conn_qp3(struct cma_id_private *id_priv, struct ibv_qp *qp)
 {
 	struct ibv_qp_attr qp_attr;
 	int ret;
 
-	ret = ucma_find_pkey(id_priv->cma_dev, id_priv->id.port_num,
-			     id_priv->id.route.addr.addr.ibaddr.pkey,
-			     &qp_attr.pkey_index);
-	if (ret)
-		return ret;
+	ret = ibv_get_pkey_index(id_priv->cma_dev->verbs, id_priv->id.port_num,
+				 id_priv->id.route.addr.addr.ibaddr.pkey);
+	if (ret < 0)
+		return ERR(EINVAL);
 
+	qp_attr.pkey_index = ret;
 	qp_attr.port_num = id_priv->id.port_num;
 	qp_attr.qp_state = IBV_QPS_INIT;
 	qp_attr.qp_access_flags = 0;
@@ -1147,12 +1136,12 @@ static int ucma_init_ud_qp3(struct cma_id_private *id_priv, struct ibv_qp *qp)
 	struct ibv_qp_attr qp_attr;
 	int ret;
 
-	ret = ucma_find_pkey(id_priv->cma_dev, id_priv->id.port_num,
-			     id_priv->id.route.addr.addr.ibaddr.pkey,
-			     &qp_attr.pkey_index);
-	if (ret)
-		return ret;
+	ret = ibv_get_pkey_index(id_priv->cma_dev->verbs, id_priv->id.port_num,
+				 id_priv->id.route.addr.addr.ibaddr.pkey);
+	if (ret < 0)
+		return ERR(EINVAL);
 
+	qp_attr.pkey_index = ret;
 	qp_attr.port_num = id_priv->id.port_num;
 	qp_attr.qp_state = IBV_QPS_INIT;
 	qp_attr.qkey = RDMA_UDP_QKEY;
@@ -1684,6 +1673,9 @@ int rdma_notify(struct rdma_cm_id *id, enum ibv_event_type event)
 
 int ucma_shutdown(struct rdma_cm_id *id)
 {
+	if (!id->verbs || !id->verbs->device)
+		return ERR(EINVAL);
+
 	switch (id->verbs->device->transport_type) {
 	case IBV_TRANSPORT_IB:
 		return ucma_modify_qp_err(id);
@@ -2085,6 +2077,18 @@ static void ucma_copy_ud_event(struct cma_event *event,
 	dst->qkey = src->qkey;
 }
 
+int rdma_establish(struct rdma_cm_id *id)
+{
+	if (id->qp)
+		return ERR(EINVAL);
+
+	/* id->qp is NULL, so ucma_process_conn_resp() will only send ACCEPT to
+	 * the passive side, and will not attempt to modify the QP.
+	 */
+	return ucma_process_conn_resp(container_of(id, struct cma_id_private,
+						   id));
+}
+
 int rdma_get_cm_event(struct rdma_event_channel *channel,
 		      struct rdma_cm_event **event)
 {
@@ -2161,12 +2165,17 @@ retry:
 		break;
 	case RDMA_CM_EVENT_CONNECT_RESPONSE:
 		ucma_copy_conn_event(evt, &resp.param.conn);
-		evt->event.status = ucma_process_conn_resp(evt->id_priv);
-		if (!evt->event.status)
-			evt->event.event = RDMA_CM_EVENT_ESTABLISHED;
-		else {
-			evt->event.event = RDMA_CM_EVENT_CONNECT_ERROR;
-			evt->id_priv->connect_error = 1;
+		if (!evt->id_priv->id.qp) {
+			evt->event.event = RDMA_CM_EVENT_CONNECT_RESPONSE;
+		} else {
+			evt->event.status =
+				ucma_process_conn_resp(evt->id_priv);
+			if (!evt->event.status)
+				evt->event.event = RDMA_CM_EVENT_ESTABLISHED;
+			else {
+				evt->event.event = RDMA_CM_EVENT_CONNECT_ERROR;
+				evt->id_priv->connect_error = 1;
+			}
 		}
 		break;
 	case RDMA_CM_EVENT_ESTABLISHED:
