@@ -127,7 +127,7 @@ struct ibv_pd *nes_ualloc_pd(struct ibv_context *context)
 		return NULL;
 	}
 	nesupd->pd_id = resp.pd_id;
-	nesupd->db_index = resp.db_index;
+	nesupd->db_index = resp.mmap_db_index;
 
 	nesupd->udoorbell = mmap(NULL, page_size, PROT_WRITE | PROT_READ, MAP_SHARED,
 			context->cmd_fd, nesupd->db_index * page_size);
@@ -168,39 +168,39 @@ int nes_ufree_pd(struct ibv_pd *pd)
 struct ibv_mr *nes_ureg_mr(struct ibv_pd *pd, void *addr,
 		size_t length, int access)
 {
-	struct ibv_mr *mr;
+	struct verbs_mr *vmr;
 	struct nes_ureg_mr cmd;
 	struct ib_uverbs_reg_mr_resp resp;
 
-	mr = malloc(sizeof *mr);
-	if (!mr)
+	vmr = malloc(sizeof(*vmr));
+	if (!vmr)
 		return NULL;
 
-	cmd.reg_type = NES_UMEMREG_TYPE_MEM;
+	cmd.reg_type = IWNES_MEMREG_TYPE_MEM;
 	if (ibv_cmd_reg_mr(pd, addr, length, (uintptr_t) addr,
-			access, mr, &cmd.ibv_cmd, sizeof cmd,
-			&resp, sizeof resp)) {
-		free(mr);
+			access, vmr, &cmd.ibv_cmd, sizeof(cmd),
+			&resp, sizeof(resp))) {
+		free(vmr);
 
 		return NULL;
 	}
 
-	return mr;
+	return &vmr->ibv_mr;
 }
 
 
 /**
  * nes_udereg_mr
  */
-int nes_udereg_mr(struct ibv_mr *mr)
+int nes_udereg_mr(struct verbs_mr *vmr)
 {
 	int ret;
 
-	ret = ibv_cmd_dereg_mr(mr);
+	ret = ibv_cmd_dereg_mr(vmr);
 	if (ret)
 		return ret;
 
-	free(mr);
+	free(vmr);
 	return 0;
 }
 
@@ -239,13 +239,13 @@ struct ibv_cq *nes_ucreate_cq(struct ibv_context *context, int cqe,
 		goto err;
 
 	/* Register the memory for the CQ */
-	reg_mr_cmd.reg_type = NES_UMEMREG_TYPE_CQ;
+	reg_mr_cmd.reg_type = IWNES_MEMREG_TYPE_CQ;
 
 	ret = ibv_cmd_reg_mr(&nesvctx->nesupd->ibv_pd, (void *)nesucq->cqes,
 			(nesucq->size*sizeof(struct nes_hw_cqe)),
-			(uintptr_t)nesucq->cqes, IBV_ACCESS_LOCAL_WRITE, &nesucq->mr,
-			&reg_mr_cmd.ibv_cmd, sizeof reg_mr_cmd,
-			&reg_mr_resp, sizeof reg_mr_resp);
+			(uintptr_t)nesucq->cqes, IBV_ACCESS_LOCAL_WRITE,
+			&nesucq->vmr, &reg_mr_cmd.ibv_cmd, sizeof(reg_mr_cmd),
+			&reg_mr_resp, sizeof(reg_mr_resp));
 	if (ret) {
 		/* fprintf(stderr, "ibv_cmd_reg_mr failed (ret = %d).\n", ret); */
 		free((struct nes_hw_cqe *)nesucq->cqes);
@@ -300,7 +300,7 @@ int nes_udestroy_cq(struct ibv_cq *cq)
 	if (ret)
 		return ret;
 
-	ret = ibv_cmd_dereg_mr(&nesucq->mr);
+	ret = ibv_cmd_dereg_mr(&nesucq->vmr);
 	if (ret)
 		fprintf(stderr, PFX "%s: Failed to deregister CQ Memory Region.\n", __FUNCTION__);
 
@@ -951,15 +951,16 @@ static int nes_vmapped_qp(struct nes_uqp *nesuqp, struct ibv_pd *pd, struct ibv_
 	nesuqp->rq_vbase = (struct nes_hw_qp_wqe *) (((char *) nesuqp->sq_vbase) +
 			   (nesuqp->sq_size * sizeof(struct nes_hw_qp_wqe)));
 
-	reg_mr_cmd.reg_type = NES_UMEMREG_TYPE_QP;
+	reg_mr_cmd.reg_type = IWNES_MEMREG_TYPE_QP;
 
 	//fprintf(stderr, PFX "qp_rq_vbase = %p qp_sq_vbase=%p reg_mr = %p\n",
 	//		nesuqp->rq_vbase, nesuqp->sq_vbase, &nesuqp->mr);
 
         ret = ibv_cmd_reg_mr(pd, (void *)nesuqp->sq_vbase,totalqpsize,
-			     (uintptr_t) nesuqp->sq_vbase, IBV_ACCESS_LOCAL_WRITE,
-			     &nesuqp->mr, &reg_mr_cmd.ibv_cmd, sizeof reg_mr_cmd,
-			     &reg_mr_resp, sizeof reg_mr_resp);
+			     (uintptr_t)nesuqp->sq_vbase,
+			     IBV_ACCESS_LOCAL_WRITE, &nesuqp->vmr,
+			     &reg_mr_cmd.ibv_cmd, sizeof(reg_mr_cmd),
+			     &reg_mr_resp, sizeof(reg_mr_resp));
         if (ret) {
                 // fprintf(stderr, PFX "%s ibv_cmd_reg_mr failed (ret = %d).\n", __FUNCTION__, ret);
 		free((void *) nesuqp->sq_vbase);
@@ -967,12 +968,12 @@ static int nes_vmapped_qp(struct nes_uqp *nesuqp, struct ibv_pd *pd, struct ibv_
         }
 	// So now the memory has been registered..
 	memset (&cmd, 0, sizeof(cmd) );
-	cmd.user_sq_buffer = (__u64) ((uintptr_t) nesuqp->sq_vbase);
+	cmd.user_wqe_buffers = (__u64) ((uintptr_t) nesuqp->sq_vbase);
 	cmd.user_qp_buffer = (__u64) ((uintptr_t) nesuqp);
 	ret = ibv_cmd_create_qp(pd, &nesuqp->ibv_qp, attr, &cmd.ibv_cmd, sizeof cmd,
 				&resp->ibv_resp, sizeof (struct nes_ucreate_qp_resp) );
 	if (ret) {
-		ibv_cmd_dereg_mr(&nesuqp->mr);
+		ibv_cmd_dereg_mr(&nesuqp->vmr);
 		free((void *)nesuqp->sq_vbase);
 		return 0;
 	}
@@ -1155,7 +1156,7 @@ int nes_udestroy_qp(struct ibv_qp *qp)
 	// fprintf(stderr, PFX "%s addr&mr= %p  \n", __FUNCTION__, &nesuqp->mr );
 
 	if (nesuqp->mapping == NES_QP_VMAP) {
-		ret = ibv_cmd_dereg_mr(&nesuqp->mr);
+		ret = ibv_cmd_dereg_mr(&nesuqp->vmr);
 		if (ret)
 	 		fprintf(stderr, PFX "%s dereg_mr FAILED\n", __FUNCTION__);
 		free((void *)nesuqp->sq_vbase);
@@ -1516,7 +1517,8 @@ int nes_udetach_mcast(struct ibv_qp *qp, const union ibv_gid *gid, uint16_t lid)
 /**
  * nes_async_event
  */
-void nes_async_event(struct ibv_async_event *event)
+void nes_async_event(struct ibv_context *context,
+		     struct ibv_async_event *event)
 {
 	struct nes_uqp *nesuqp;
 
