@@ -125,35 +125,34 @@ static int rxe_dealloc_pd(struct ibv_pd *pd)
 static struct ibv_mr *rxe_reg_mr(struct ibv_pd *pd, void *addr, size_t length,
 				 int access)
 {
-	struct ibv_mr *mr;
+	struct verbs_mr *vmr;
 	struct ibv_reg_mr cmd;
 	struct ib_uverbs_reg_mr_resp resp;
 	int ret;
 
-	mr = malloc(sizeof *mr);
-	if (!mr) {
+	vmr = malloc(sizeof(*vmr));
+	if (!vmr)
 		return NULL;
-	}
 
-	ret = ibv_cmd_reg_mr(pd, addr, length, (uintptr_t)addr, access, mr,
+	ret = ibv_cmd_reg_mr(pd, addr, length, (uintptr_t)addr, access, vmr,
 			     &cmd, sizeof cmd, &resp, sizeof resp);
 	if (ret) {
-		free(mr);
+		free(vmr);
 		return NULL;
 	}
 
-	return mr;
+	return &vmr->ibv_mr;
 }
 
-static int rxe_dereg_mr(struct ibv_mr *mr)
+static int rxe_dereg_mr(struct verbs_mr *vmr)
 {
 	int ret;
 
-	ret = ibv_cmd_dereg_mr(mr);
+	ret = ibv_cmd_dereg_mr(vmr);
 	if (ret)
 		return ret;
 
-	free(mr);
+	free(vmr);
 	return 0;
 }
 
@@ -162,8 +161,7 @@ static struct ibv_cq *rxe_create_cq(struct ibv_context *context, int cqe,
 				    int comp_vector)
 {
 	struct rxe_cq *cq;
-	struct ibv_create_cq cmd;
-	struct rxe_create_cq_resp resp;
+	struct urxe_create_cq_resp resp;
 	int ret;
 
 	cq = malloc(sizeof *cq);
@@ -172,7 +170,7 @@ static struct ibv_cq *rxe_create_cq(struct ibv_context *context, int cqe,
 	}
 
 	ret = ibv_cmd_create_cq(context, cqe, channel, comp_vector,
-				&cq->ibv_cq, &cmd, sizeof cmd,
+				&cq->ibv_cq, NULL, 0,
 				&resp.ibv_resp, sizeof resp);
 	if (ret) {
 		free(cq);
@@ -197,7 +195,7 @@ static int rxe_resize_cq(struct ibv_cq *ibcq, int cqe)
 {
 	struct rxe_cq *cq = to_rcq(ibcq);
 	struct ibv_resize_cq cmd;
-	struct rxe_resize_cq_resp resp;
+	struct urxe_resize_cq_resp resp;
 	int ret;
 
 	pthread_spin_lock(&cq->lock);
@@ -274,7 +272,7 @@ static struct ibv_srq *rxe_create_srq(struct ibv_pd *pd,
 {
 	struct rxe_srq *srq;
 	struct ibv_create_srq cmd;
-	struct rxe_create_srq_resp resp;
+	struct urxe_create_srq_resp resp;
 	int ret;
 
 	srq = malloc(sizeof *srq);
@@ -309,9 +307,9 @@ static int rxe_modify_srq(struct ibv_srq *ibsrq,
 		   struct ibv_srq_attr *attr, int attr_mask)
 {
 	struct rxe_srq *srq = to_rsrq(ibsrq);
-	struct rxe_modify_srq_cmd cmd;
+	struct urxe_modify_srq cmd;
 	int rc = 0;
-	struct mmap_info mi;
+	struct mminfo mi;
 
 	mi.offset = 0;
 	mi.size = 0;
@@ -440,7 +438,7 @@ static struct ibv_qp *rxe_create_qp(struct ibv_pd *pd,
 				    struct ibv_qp_init_attr *attr)
 {
 	struct ibv_create_qp cmd;
-	struct rxe_create_qp_resp resp;
+	struct urxe_create_qp_resp resp;
 	struct rxe_qp *qp;
 	int ret;
 
@@ -587,6 +585,7 @@ static void convert_send_wr(struct rxe_send_wr *kwr, struct ibv_send_wr *uwr)
 	case IBV_WR_BIND_MW:
 	case IBV_WR_SEND_WITH_INV:
 	case IBV_WR_TSO:
+	case IBV_WR_DRIVER1:
 		break;
 	}
 }
@@ -804,6 +803,10 @@ static struct ibv_ah *rxe_create_ah(struct ibv_pd *pd, struct ibv_ah_attr *attr)
 
 	rdma_gid2ip(&av->sgid_addr, &sgid);
 	rdma_gid2ip(&av->dgid_addr, &attr->grh.dgid);
+	if (ibv_resolve_eth_l2_from_gid(pd->context, attr, av->dmac, NULL)) {
+		free(ah);
+		return NULL;
+	}
 
 	memset(&resp, 0, sizeof(resp));
 	if (ibv_cmd_create_ah(pd, &ah->ibv_ah, attr, &resp, sizeof(resp))) {
@@ -857,13 +860,15 @@ static const struct verbs_context_ops rxe_ctx_ops = {
 };
 
 static struct verbs_context *rxe_alloc_context(struct ibv_device *ibdev,
-					       int cmd_fd)
+					       int cmd_fd,
+					       void *private_data)
 {
 	struct rxe_context *context;
 	struct ibv_get_context cmd;
 	struct ib_uverbs_get_context_resp resp;
 
-	context = verbs_init_and_alloc_context(ibdev, cmd_fd, context, ibv_ctx);
+	context = verbs_init_and_alloc_context(ibdev, cmd_fd, context, ibv_ctx,
+					       RDMA_DRIVER_RXE);
 	if (!context)
 		return NULL;
 
@@ -910,12 +915,17 @@ static struct verbs_device *rxe_device_alloc(struct verbs_sysfs_dev *sysfs_dev)
 
 static const struct verbs_device_ops rxe_dev_ops = {
 	.name = "rxe",
-	.match_min_abi_version = 0,
-	.match_max_abi_version = INT_MAX,
+	/*
+	 * For 64 bit machines ABI version 1 and 2 are the same. Otherwise 32
+	 * bit machines require ABI version 2 which guarentees the user and
+	 * kernel use the same ABI.
+	 */
+	.match_min_abi_version = sizeof(void *) == 8?1:2,
+	.match_max_abi_version = 2,
 	.match_table = hca_table,
 	.alloc_device = rxe_device_alloc,
 	.uninit_device = rxe_uninit_device,
 	.alloc_context = rxe_alloc_context,
 	.free_context = rxe_free_context,
 };
-PROVIDER_DRIVER(rxe_dev_ops);
+PROVIDER_DRIVER(rxe, rxe_dev_ops);
