@@ -401,7 +401,7 @@ static int is_enabled_by_rules_file(struct target_details *target)
 
 static int add_non_exist_target(struct target_details *target)
 {
-	char scsi_host_dir[50];
+	char scsi_host_dir[256];
 	DIR *dir;
 	struct dirent *subdir;
 	char *subdir_name_ptr;
@@ -679,8 +679,7 @@ static int translate_umad_to_ibdev_and_port(char *umad_dev, char **ibdev,
 
 	umad_dev_name = rindex(umad_dev, '/');
 	if (!umad_dev_name) {
-		pr_err("Couldn't find device name in '%s'\n",
-			umad_dev_name);
+		pr_err("Couldn't find device name in '%s'\n", umad_dev);
 		return -1;
 	}
 
@@ -793,7 +792,29 @@ static int check_sm_cap(struct umad_resources *umad_res, int *mask_match)
 	return 0;
 }
 
-static int set_class_port_info(struct umad_resources *umad_res, uint16_t dlid)
+int pkey_index_to_pkey(struct umad_resources *umad_res, int pkey_index,
+		       __be16 *pkey)
+{
+	if (ibv_query_pkey(umad_res->ib_ctx, config->port_num, pkey_index,
+			   pkey) < 0)
+		return -1;
+	if (*pkey)
+		pr_debug("discover Targets for P_key %04x (index %d)\n",
+			 *pkey, pkey_index);
+	return 0;
+}
+
+static int pkey_to_pkey_index(struct umad_resources *umad_res, uint16_t h_pkey,
+			      uint16_t *pkey_index)
+{
+	int res = ibv_get_pkey_index(umad_res->ib_ctx, config->port_num,
+				     htobe16(h_pkey));
+	if (res >= 0)
+		*pkey_index = res;
+	return res;
+}
+
+static int set_class_port_info(struct umad_resources *umad_res, uint16_t dlid, uint16_t h_pkey)
 {
 	struct srp_ib_user_mad		in_mad, out_mad;
 	struct umad_dm_packet	       *out_dm_mad, *in_dm_mad;
@@ -803,6 +824,12 @@ static int set_class_port_info(struct umad_resources *umad_res, uint16_t dlid)
 
 	init_srp_dm_mad(&out_mad, umad_res->agent, dlid, UMAD_ATTR_CLASS_PORT_INFO, 0);
 
+	if (pkey_to_pkey_index(umad_res, h_pkey, &out_mad.hdr.addr.pkey_index)
+	    < 0) {
+		pr_err("set_class_port_info: Unable to find pkey_index for pkey %#x\n", h_pkey);
+		return -1;
+	}
+
 	out_dm_mad = get_data_ptr(out_mad);
 	out_dm_mad->mad_hdr.method = UMAD_METHOD_SET;
 
@@ -837,12 +864,17 @@ static int set_class_port_info(struct umad_resources *umad_res, uint16_t dlid)
 }
 
 static int get_iou_info(struct umad_resources *umad_res, uint16_t dlid,
-			struct srp_dm_iou_info *iou_info)
+			uint16_t h_pkey, struct srp_dm_iou_info *iou_info)
 {
 	struct srp_ib_user_mad		in_mad, out_mad;
 	struct umad_dm_packet	       *in_dm_mad;
 
 	init_srp_dm_mad(&out_mad, umad_res->agent, dlid, SRP_DM_ATTR_IO_UNIT_INFO, 0);
+	if (pkey_to_pkey_index(umad_res, h_pkey, &out_mad.hdr.addr.pkey_index)
+	    < 0) {
+		pr_err("get_iou_info: Unable to find pkey_index for pkey %#x\n", h_pkey);
+		return -1;
+	}
 
 	if (send_and_get(umad_res->portid, umad_res->agent, &out_mad, &in_mad, 0) < 0)
 		return -1;
@@ -861,7 +893,7 @@ static int get_iou_info(struct umad_resources *umad_res, uint16_t dlid,
 	return 0;
 }
 
-static int get_ioc_prof(struct umad_resources *umad_res, uint16_t h_dlid, int ioc,
+static int get_ioc_prof(struct umad_resources *umad_res, uint16_t h_dlid, uint16_t h_pkey, int ioc,
 			struct srp_dm_ioc_prof *ioc_prof)
 {
 	struct srp_ib_user_mad		in_mad, out_mad;
@@ -869,6 +901,13 @@ static int get_ioc_prof(struct umad_resources *umad_res, uint16_t h_dlid, int io
 
 	init_srp_dm_mad(&out_mad, umad_res->agent, h_dlid, SRP_DM_ATTR_IO_CONTROLLER_PROFILE, ioc);
 
+	if (pkey_to_pkey_index(umad_res, h_pkey, &out_mad.hdr.addr.pkey_index)
+	    < 0) {
+		pr_err("get_ioc_prof: Unable to find pkey_index for pkey %#x\n",
+		       h_pkey);
+		return -1;
+	}
+
 	if (send_and_get(umad_res->portid, umad_res->agent, &out_mad, &in_mad, 0) < 0)
 		return -1;
 
@@ -884,7 +923,7 @@ static int get_ioc_prof(struct umad_resources *umad_res, uint16_t h_dlid, int io
 	return 0;
 }
 
-static int get_svc_entries(struct umad_resources *umad_res, uint16_t dlid, int ioc,
+static int get_svc_entries(struct umad_resources *umad_res, uint16_t dlid, uint16_t h_pkey, int ioc,
 			   int start, int end, struct srp_dm_svc_entries *svc_entries)
 {
 	struct srp_ib_user_mad		in_mad, out_mad;
@@ -893,6 +932,13 @@ static int get_svc_entries(struct umad_resources *umad_res, uint16_t dlid, int i
 	init_srp_dm_mad(&out_mad, umad_res->agent, dlid, SRP_DM_ATTR_SERVICE_ENTRIES,
 			(ioc << 16) | (end << 8) | start);
 
+	if (pkey_to_pkey_index(umad_res, h_pkey, &out_mad.hdr.addr.pkey_index)
+	    < 0) {
+		pr_err("get_svc_entries: Unable to find pkey_index for pkey %#x\n",
+		       h_pkey);
+		return -1;
+	}
+
 	if (send_and_get(umad_res->portid, umad_res->agent, &out_mad, &in_mad, 0) < 0)
 		return -1;
 
@@ -928,10 +974,10 @@ static int do_port(struct resources *res, uint16_t pkey, uint16_t dlid,
 
  	pr_debug("enter do_port\n");
 	if ((target->h_guid & oui_mask) == topspin_oui &&
-	    set_class_port_info(umad_res, dlid))
+	    set_class_port_info(umad_res, dlid, pkey))
 		pr_err("Warning: set of ClassPortInfo failed\n");
 
-	ret = get_iou_info(umad_res, dlid, &iou_info);
+	ret = get_iou_info(umad_res, dlid, pkey, &iou_info);
 	if (ret < 0) {
 		pr_err("failed to get iou info for dlid %#x\n", dlid);
 		goto out;
@@ -962,7 +1008,7 @@ static int do_port(struct resources *res, uint16_t pkey, uint16_t dlid,
 		    SRP_DM_IOC_PRESENT) {
 			pr_human("\n");
 
-			if (get_ioc_prof(umad_res, dlid, i + 1, &target->ioc_prof))
+			if (get_ioc_prof(umad_res, dlid, pkey, i + 1, &target->ioc_prof))
 				continue;
 
 			pr_human("    controller[%3d]\n", i + 1);
@@ -982,7 +1028,7 @@ static int do_port(struct resources *res, uint16_t pkey, uint16_t dlid,
 				if (n >= target->ioc_prof.service_entries)
 					n = target->ioc_prof.service_entries - 1;
 
-				if (get_svc_entries(umad_res, dlid, i + 1,
+				if (get_svc_entries(umad_res, dlid, pkey, i + 1,
 						    j, n, &svc_entries))
 					continue;
 
@@ -1071,24 +1117,6 @@ static int get_port_info(struct umad_resources *umad_res, uint16_t dlid,
 	return 0;
 }
 
-int pkey_index_to_pkey(struct umad_resources *umad_res, int pkey_index,
-		       uint16_t *pkey)
-{
-	char pkey_file[16], pkey_str[16];
-
-	/* Read pkey */
-	snprintf(pkey_file, sizeof(pkey_file), "pkeys/%d", pkey_index);
-	if (srpd_sys_read_string(umad_res->port_sysfs_path, pkey_file,
-				 pkey_str, sizeof(pkey_str)) < 0)
-		return -1;
-
-	*pkey = strtoul(pkey_str, NULL, 0);
-	if (*pkey)
-		pr_debug("discover Targets for P_key %04x (index %d)\n",
-			 *pkey, pkey_index);
-	return 0;
-}
-
 static int get_shared_pkeys(struct resources *res,
 			    uint16_t dest_port_lid,
 			    uint16_t *pkeys)
@@ -1101,7 +1129,7 @@ static int get_shared_pkeys(struct resources *res,
 	struct ib_path_rec	       *path_rec;
 	ssize_t len;
 	int i, num_pkeys = 0;
-	uint16_t pkey;
+	__be16 pkey;
 	uint16_t local_port_lid = get_port_lid(res->ud_res->ib_ctx,
 					       config->port_num, NULL);
 
@@ -1135,7 +1163,7 @@ static int get_shared_pkeys(struct resources *res,
 		path_rec = (struct ib_path_rec *)out_sa_mad->data;
 		path_rec->slid = htobe16(local_port_lid);
 		path_rec->dlid = htobe16(dest_port_lid);
-		path_rec->pkey = htobe16(pkey);
+		path_rec->pkey = pkey;
 
 		len = send_and_get(umad_res->portid, umad_res->agent, &out_mad,
 				   (struct srp_ib_user_mad *)in_mad,
@@ -1934,6 +1962,7 @@ static struct resources *alloc_res(void)
 	if (ret)
 		goto err;
 	res->res.ud_res = &res->ud_res;
+	res->umad_res.ib_ctx = res->ud_res.ib_ctx;
 
 	ret = sync_resources_init(&res->sync_res);
 	if (ret)
@@ -2103,8 +2132,12 @@ int main(int argc, char *argv[])
 	BUILD_ASSERT(sizeof(struct ib_path_rec) == 64);
 	BUILD_ASSERT(sizeof(struct ib_inform_info) == 36);
 	BUILD_ASSERT(sizeof(struct ib_mad_notice_attr) == 80);
+	BUILD_ASSERT(offsetof(struct ib_mad_notice_attr, generic.trap_num) ==
+		     4);
+	BUILD_ASSERT(offsetof(struct ib_mad_notice_attr, vend.dev_id) == 4);
+	BUILD_ASSERT(offsetof(struct ib_mad_notice_attr, ntc_64_67.gid) == 16);
 	BUILD_ASSERT(offsetof(struct ib_mad_notice_attr,
-			      data_details.ntc_64_67.gid) == 16);
+			      ntc_144.new_cap_mask) == 16);
 #endif
 	BUILD_ASSERT(sizeof(struct srp_sa_node_rec) == 108);
 	BUILD_ASSERT(sizeof(struct srp_sa_port_info_rec) == 58);
@@ -2377,7 +2410,7 @@ static int get_lid(struct umad_resources *umad_res, union umad_gid *gid,
 
 	path_rec->sgid = *gid;
 	path_rec->dgid = *gid;
-	path_rec->num_path = 1;
+	path_rec->reversible_numpath = 1;
 	path_rec->hop_flow_raw = htobe32(1 << 31); /* rawtraffic=1 hoplimit = 0 */
 
 	if (send_and_get(umad_res->portid, umad_res->agent, &out_mad, &in_mad, 0) < 0)
